#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import numpy as np

try:
    import vedo.vtkclasses as vtk
except ImportError:
    import vtkmodules.all as vtk

import vedo
from vedo import settings
from vedo import utils
from vedo.pointcloud import Points
from vedo.mesh import Mesh
from vedo.volume import Volume

__doc__ = """Submodule to delegate jupyter notebook rendering"""

__all__ = []


############################################################################################
def get_notebook_backend(actors2show=()):
    """Return the appropriate notebook viewer"""

    #########################################
    if settings.default_backend == "2d":
        return start_2d()

    #########################################
    if settings.default_backend == "k3d":
        return start_k3d(actors2show)

    #########################################
    if settings.default_backend.startswith("trame"):
        return start_trame()

    #########################################
    if settings.default_backend.startswith("ipyvtk"):
        return start_ipyvtklink()

    vedo.logger.error(f"Unknown jupyter backend: {settings.default_backend}")
    return None


#####################################################################################
def start_2d():
    try:
        import PIL.Image
        # import IPython
    except ImportError:
        print("PIL or IPython not available")
        return None

    plt = vedo.plotter_instance

    if hasattr(plt, "window") and plt.window:
        if plt.renderer == plt.renderers[-1]:
            nn = vedo.file_io.screenshot(asarray=True, scale=1)
            pil_img = PIL.Image.fromarray(nn)
            # IPython.display.display(pil_img)
            vedo.notebook_plotter = pil_img
            if settings.backend_autoclose:
                plt.close()
            return pil_img


####################################################################################
def start_k3d(actors2show):

    try:
        # https://github.com/K3D-tools/K3D-jupyter
        import k3d
    except ModuleNotFoundError:
        print("Cannot find k3d, install with:  pip install k3d")
        return None

    plt = vedo.plotter_instance
    if not plt:
        return None

    actors2show2 = []
    for ia in actors2show:
        if not ia:
            continue
        if isinstance(ia, vedo.Assembly):  # unpack assemblies
            actors2show2 += ia.recursive_unpack()
        else:
            actors2show2.append(ia)

    vedo.notebook_plotter = k3d.plot(
        axes=["x", "y", "z"],
        menu_visibility=settings.k3d_menu_visibility,
        height=settings.k3d_plot_height,
        antialias=settings.k3d_antialias,
        background_color=_rgb2int(vedo.get_color(plt._bg)),
        camera_fov=30.0,  # deg (this is the vtk default)
        lighting=settings.k3d_lighting,
        grid_color=_rgb2int(vedo.get_color(settings.k3d_axes_color)),
        label_color=_rgb2int(vedo.get_color(settings.k3d_axes_color)),
        axes_helper=settings.k3d_axes_helper,
        # axes_helper_colors=[_rgb2int(vedo.get_color("red5")), # not working
        #                     _rgb2int(vedo.get_color("green5")),
        #                     _rgb2int(vedo.get_color("blue5"))],
    )
    # vedo.notebook_plotter.axes_helper_colors = [
    #     vedo.backends._rgb2int(vedo.get_color("red5")), # not working
    #     vedo.backends._rgb2int(vedo.get_color("green5")),
    #     vedo.backends._rgb2int(vedo.get_color("blue5"))
    # ]

    # set k3d camera
    vedo.notebook_plotter.camera_auto_fit = settings.k3d_camera_autofit
    vedo.notebook_plotter.grid_auto_fit = settings.k3d_grid_autofit
    vedo.notebook_plotter.axes_helper = settings.k3d_axes_helper

    if plt.camera:
        vedo.notebook_plotter.camera = utils.vtkCameraToK3D(plt.camera)

    if not plt.axes:
        vedo.notebook_plotter.grid_visible = False

    for ia in actors2show2:

        if isinstance(ia, (vtk.vtkCornerAnnotation,
                           vtk.vtkAssembly,
                           vtk.vtkActor2D)):
            continue

        iacloned = ia

        kobj = None
        kcmap = None
        color_attribute = None
        vtkscals = None
        name = None
        if hasattr(ia, "filename"):
            if ia.filename:
                name = os.path.basename(ia.filename)
            if ia.name:
                name = os.path.basename(ia.name)

        ################################################################## scalars
        # work out scalars first, Points Lines are also Mesh objs
        if isinstance(ia, Points):
            # print('scalars', ia.name, ia.npoints)
            iap = ia.GetProperty()

            if ia.inputdata().GetNumberOfPolys():
                iacloned = ia.clone()
                iapoly = iacloned.clean().triangulate().compute_normals().polydata()
            else:
                iapoly = ia.polydata()

            if ia.mapper().GetScalarVisibility() and ia.mapper().GetColorMode() > 0:

                vtkdata = iapoly.GetPointData()
                vtkscals = vtkdata.GetScalars()

                if vtkscals is None:
                    vtkdata = iapoly.GetCellData()
                    vtkscals = vtkdata.GetScalars()
                    if vtkscals is not None:
                        c2p = vtk.vtkCellDataToPointData()
                        c2p.SetInputData(iapoly)
                        c2p.Update()
                        iapoly = c2p.GetOutput()
                        vtkdata = iapoly.GetPointData()
                        vtkscals = vtkdata.GetScalars()

                else:

                    if not vtkscals.GetName():
                        vtkscals.SetName("scalars")
                    scals_min, scals_max = ia.mapper().GetScalarRange()
                    color_attribute = (vtkscals.GetName(), scals_min, scals_max)
                    lut = ia.mapper().GetLookupTable()
                    lut.Build()
                    kcmap = []
                    nlut = lut.GetNumberOfTableValues()
                    for i in range(nlut):
                        r, g, b, _ = lut.GetTableValue(i)
                        kcmap += [i / (nlut - 1), r, g, b]

            else:
                color_attribute = ia.color()

        #####################################################################Volume
        if isinstance(ia, Volume):
            # print('Volume', ia.name, ia.dimensions())
            kx, ky, _ = ia.dimensions()
            arr = ia.pointdata[0]
            kimage = arr.reshape(-1, ky, kx)

            colorTransferFunction = ia.GetProperty().GetRGBTransferFunction()
            kcmap = []
            for i in range(128):
                r, g, b = colorTransferFunction.GetColor(i / 127)
                kcmap += [i / 127, r, g, b]

            kbounds = np.array(ia.imagedata().GetBounds()) + np.repeat(
                np.array(ia.imagedata().GetSpacing()) / 2.0, 2
            ) * np.array([-1, 1] * 3)

            kobj = k3d.volume(
                kimage.astype(np.float32),
                color_map=kcmap,
                # color_range=ia.imagedata().GetScalarRange(),
                alpha_coef=10,
                bounds=kbounds,
                name=name,
            )
            vedo.notebook_plotter += kobj

        ################################################################ Text2D
        elif isinstance(ia, vedo.Text2D):
            # print('Text2D', ia.GetPosition())
            pos = (ia.GetPosition()[0], 1.0 - ia.GetPosition()[1])

            kobj = k3d.text2d(
                ia.text(),
                position=pos,
                color=_rgb2int(vedo.get_color(ia.c())),
                is_html=True,
                size=ia.property.GetFontSize() / 22.5 * 1.5,
                label_box=bool(ia.property.GetFrame()),
                # reference_point='bl',
            )
            vedo.notebook_plotter += kobj

        ################################################################# Lines
        elif (hasattr(ia, "polydata") 
              and ia.polydata(False).GetNumberOfLines() 
              and ia.polydata(False).GetNumberOfPolys() == 0):

            for i, ln_idx in enumerate(ia.lines()):

                if i > 200:
                    vedo.logger.warning("in k3d, nr. of lines is limited to 200.")
                    break

                pts = ia.points()[ln_idx]

                aves = ia.diagonal_size() * iap.GetLineWidth() / 100

                kobj = k3d.line(
                    pts.astype(np.float32),
                    color=_rgb2int(iap.GetColor()),
                    opacity=iap.GetOpacity(),
                    shader=settings.k3d_line_shader,
                    width=aves,
                    name=name,
                )
                vedo.notebook_plotter += kobj

        ################################################################## Mesh
        elif isinstance(ia, Mesh) and ia.npoints and ia.polydata(False).GetNumberOfPolys():
            # print('Mesh', ia.name, ia.npoints, len(ia.faces()))

            if not vtkscals:
                color_attribute = None

            cols = []
            if ia.mapper().GetColorMode() == 0:
                # direct RGB colors

                vcols = ia.inputdata().GetPointData().GetScalars()
                if vcols and vcols.GetNumberOfComponents() == 3:
                    cols = utils.vtk2numpy(vcols)
                    cols = 65536 * cols[:, 0] + 256 * cols[:, 1] + cols[:, 2]
                # print("GetColor",iap.GetColor(), _rgb2int(iap.GetColor()) )
                # print("colors", len(cols))
                # print("color_attribute", color_attribute)
                # if kcmap is not None: print("color_map", len(kcmap))
                # TODO:
                # https://k3d-jupyter.org/reference/factory.mesh.html#colormap

                kobj = k3d.mesh(
                    iacloned.points(),
                    iacloned.faces(),
                    colors=cols,
                    name=name,
                    color=_rgb2int(iap.GetColor()),
                    opacity=iap.GetOpacity(),
                    side="double",
                    wireframe=(iap.GetRepresentation() == 1),
                )

            else:

                kobj = k3d.vtk_poly_data(
                    iapoly,
                    name=name,
                    color=_rgb2int(iap.GetColor()),
                    color_attribute=color_attribute,
                    color_map=kcmap,
                    opacity=iap.GetOpacity(),
                    side="double",
                    wireframe=(iap.GetRepresentation() == 1),
                )

            if iap.GetInterpolation() == 0:
                kobj.flat_shading = True

            vedo.notebook_plotter += kobj

        #####################################################################Points
        elif isinstance(ia, Points):
            # print('Points', ia.name, ia.npoints)
            kcols = []
            if kcmap is not None and vtkscals:
                scals = utils.vtk2numpy(vtkscals)
                kcols = k3d.helpers.map_colors(scals, kcmap, [scals_min, scals_max]).astype(
                    np.uint32
                )

            aves = ia.average_size() * iap.GetPointSize() / 200

            kobj = k3d.points(
                ia.points().astype(np.float32),
                color=_rgb2int(iap.GetColor()),
                colors=kcols,
                opacity=iap.GetOpacity(),
                shader=settings.k3d_point_shader,
                point_size=aves,
                name=name,
            )
            vedo.notebook_plotter += kobj

        #####################################################################
        elif isinstance(ia, vedo.Picture):
            vedo.logger.error("Sorry Picture objects are not supported in k3d.")

    if plt and settings.backend_autoclose:
        plt.close()
    return vedo.notebook_plotter


#####################################################################################
def start_trame():

    try:
        from trame.app import get_server, jupyter
        from trame.ui.vuetify import VAppLayout
        from trame.widgets import vtk as t_vtk, vuetify
    except ImportError:
        print("trame is not installed, try:\n> pip install trame")
        return None

    plt = vedo.plotter_instance
    if hasattr(plt, "window") and plt.window:
        plt.renderer.ResetCamera()
        server = get_server("jupyter-1")
        state, ctrl = server.state, server.controller
        plt.server = server
        plt.controller = ctrl
        plt.state = state

        with VAppLayout(server) as layout:

            with layout.root:

                with vuetify.VContainer(fluid=True, classes="pa-0 fill-height"):
                    plt.reset_camera()
                    view = t_vtk.VtkLocalView(plt.window)
                    ctrl.view_update = view.update
                    ctrl.view_reset_camera = view.reset_camera

        ctrl.on_server_exited.add(lambda **_: print("trame server exited"))
        vedo.notebook_plotter = jupyter.show(server)
        return vedo.notebook_plotter
    vedo.logger.error("No window present for the trame backend.")
    return None


#####################################################################################
def start_ipyvtklink():
    try:
        from ipyvtklink.viewer import ViewInteractiveWidget
    except ImportError:
        print("ipyvtklink is not installed, try:\n> pip install ipyvtklink")
        return None

    plt = vedo.plotter_instance
    if hasattr(plt, "window") and plt.window:
        plt.renderer.ResetCamera()
        vedo.notebook_plotter = ViewInteractiveWidget(
            plt.window, allow_wheel=True, quality=100, quick_quality=50
        )
        return vedo.notebook_plotter
    vedo.logger.error("No window present for the ipyvtklink backend.")
    return None


#####################################################################################
def _rgb2int(rgb_tuple):
    # Return the int number of a color from (r,g,b), with 0<r<1 etc.
    rgb = (int(rgb_tuple[0] * 255), int(rgb_tuple[1] * 255), int(rgb_tuple[2] * 255))
    return 65536 * rgb[0] + 256 * rgb[1] + rgb[2]



"""Create the Convex Hull of a Mesh or a set of input points"""
from vedo import *

settings.default_font = 'Bongas'
settings.use_depth_peeling = True

spid = Mesh(dataurl+"spider.ply").c("brown")

ch = ConvexHull(spid.points()).alpha(0.2)

show(spid, ch, __doc__, axes=1).close()



from vedo import *

man = Mesh(dataurl+'man.vtk').c('k9').lighting('glossy')
floor = Box(length=9, width=9, height=0.1).z(-1.6).c('white')
cube = Cube().pos(2,-2,-1)

p1 = Arrow([4,0,4], [0,0,0],  c='red5').scale(0.2)
p2 = Arrow([0,4,4], [0,0,0],  c='green5').scale(0.2)
p3 = Arrow([-4,-4,4], [0,0,0],  c='blue5').scale(0.2)

# Add light sources at the given positions
# (grab the position and color of the arrow object)
l1 = Light(p1)
l2 = Light(p2)
l3 = Light(p3)

plt = Plotter(bg='blackboard').add_shadows()
plt.show(l1, l2, l3, p1, p2, p3, man, floor, cube)




"""Add a square button with N possible internal states
to a rendering window that calls an external function"""
from vedo import Plotter, Mesh, dataurl, printc

# Define a function that toggles the transparency of a mesh
#  and changes the button state
def buttonfunc():
    mesh.alpha(1 - mesh.alpha())  # toggle mesh transparency
    bu.switch()                   # change to next status
    printc(bu.status(), box="_", dim=True)

# Load a mesh and set its color to violet
mesh = Mesh(dataurl+"magnolia.vtk").c("violet").flat()

# Create an instance of the Plotter class with axes style-11 enabled
plt = Plotter(axes=11)

# Add a button to the plotter with buttonfunc as the callback function
bu = plt.add_button(
    buttonfunc,
    pos=(0.7, 0.05),  # x,y fraction from bottom left corner
    states=["click to hide", "click to show"],  # text for each state
    c=["w", "w"],     # font color for each state
    bc=["dg", "dv"],  # background color for each state
    font="courier",   # font type
    size=25,          # font size
    bold=True,        # bold font
    italic=False,     # non-italic font style
)

# Show the mesh, docstring, and button in the plot
plt.show(mesh, __doc__).close()



"""Visualize scalar values interactively
by hovering the mouse on a mesh
Press c to clear the path"""
from vedo import *

def func(evt):                       ### called every time mouse moves!
    msh = evt.actor
    if not msh:
        return                       # mouse hits nothing, return.
    pt = evt.picked3d                # 3d coords of point under mouse

    pid = msh.closest_point(pt, return_point_id=True)
    txt = f"Point:  {precision(pt[:2]  ,2)}\n" \
          f"Height: {precision(arr[pid],3)}\n" \
          f"Ground speed: {precision(evt.speed3d*100,2)}"
    msg.text(txt)                    # update text message

    arw = Arrow(pt - evt.delta3d, pt, s=0.001, c='orange5')
    fp = msh.flagpole(
        txt, point=pt, offset=(0.4,0.6),
        s=0.04, c='k', font="VictorMono",
    ).follow_camera()  # make it always face the camera
    if len(plt.actors) > 3:
        plt.pop()                    # remove the old flagpole
    plt.add(arw, fp).render()        # add Arrow and the new flagpole

msg = Text2D(pos='bottom-left', font="VictorMono") # an empty text
hil = ParametricShape('RandomHills').cmap('terrain').add_scalarbar()
arr = hil.pointdata["Scalars"]       # numpy array with heights

plt = Plotter(axes=1, bg2='lightblue')

plt.add_callback('mouse move', func) # add the callback function
plt.add_callback('keyboard', lambda evt: plt.remove(plt.actors[3:]).render())

plt.show(hil, msg, __doc__, viewup='z')
plt.close()




"""Insert 2D and 3D scalarbars
in the rendering scene"""
from vedo import Mesh, dataurl, show

shape = Mesh(dataurl + "lamp.vtk")

ms = []
cmaps = ("jet", "PuOr", "viridis")
for i in range(3):
    s = shape.clone(deep=False).pos(0, i * 2.2, 0)
    # colorize mesh
    scals = s.points()[:, 2]
    s.cmap(cmaps[i], scals)
    ms.append(s)

# add 2D scalar bar to first mesh
ms[0].add_scalarbar(title="my scalarbar\nnumber #0")  # 2D

# add 3D scalar bars
ms[1].add_scalarbar3d(c="k", title="scalarbar #1", size=[None, 3])

sc = ms[2].add_scalarbar3d(
    pos=(1, 0, -5),
    c="k",
    size=[None, 2.8],  # change y-size only
    title="A viridis 3D\nscalarbar to play with",
    title_font="Quikhand",
    title_xoffset=-2,  # offset of labels
    title_size=1.5,
)
sc.scalarbar.rotate_x(90)  # make it vertical

show(ms, __doc__, axes=1, viewup="z").close()



"""Hover mouse onto an object
to pop a flag-style label"""
from vedo import *

b = Mesh(dataurl+'bunny.obj').color('m')
c = Cube(side=0.1).compute_normals().alpha(0.8).y(-0.02).lighting("off").lw(1)

fp = b.flagpole('A flag pole descriptor\nfor a rabbit', font='Quikhand')
fp.scale(0.5).color('v').use_bounds() # tell camera to take fp bounds into account

c.caption('2d caption for a cube\nwith face indices', point=[0.044, 0.03, -0.04],
          size=(0.3,0.06), font="VictorMono", alpha=1)

# create a new object made of polygonal text labels to indicate the cell numbers
flabs = c.labels('id', on="cells", font='Theemim', scale=0.02, c='k')
vlabs = c.clone().clean().labels2d(font='ComicMono', scale=3, bc='orange7')

# create a custom entry to the legend
b.legend('Bugs the bunny')
c.legend('The Cube box')
lbox = LegendBox([b,c], font="Bongas", width=0.25)

show(b, c, fp, flabs, vlabs, lbox, __doc__, axes=11, bg2='linen').close()



"""Map an array which is defined on
the vertices of a mesh to its cells"""
from vedo import *

doc = Text2D(__doc__, pos="top-center")

mesh1 = Mesh(dataurl+'icosahedron.vtk').linewidth(0.1).flat()

# let the scalar be the z coordinate of the mesh vertices
msg1 = Text2D("Scalars originally defined on points..", pos="bottom-center")
mesh1.pointdata["myzscalars"] = mesh1.points()[:, 2]

mesh1.cmap("jet", "myzscalars", on="points")

msg2 = Text2D("..are interpolated to cells.", pos="bottom-center")
mesh2 = mesh1.clone(deep=False).map_points_to_cells()

plt = Plotter(N=2, axes=11)
plt.at(0).show(mesh1, msg1, doc, viewup="z")
plt.at(1).show(mesh2, msg2)
plt.interactive().close()



"""Embed a 3D scene
in a webpage with x3d"""
from vedo import dataurl, Plotter, Volume, Text3D

plt = Plotter(size=(800,600), bg='GhostWhite')

embryo = Volume(dataurl+'embryo.tif').isosurface().decimate(0.5)
coords = embryo.points()
embryo.cmap('PRGn', coords[:,1]) # add dummy colors along y

txt = Text3D(__doc__, font='Bongas', s=350, c='red2', depth=0.05)
txt.pos(2500, 300, 500)

plt.show(embryo, txt, txt.box(padding=250), axes=1, viewup='z', zoom=1.2)

# This exports the scene and generates 2 files:
# embryo.x3d and an example embryo.html to inspect in the browser
plt.export('embryo.x3d', binary=False)

print("Type: \n firefox embryo.html")



"""Colorize faces of a Mesh by passing
a 1-to-1 list of colors and transparencies"""
from vedo import *

# Enable depth peeling for better rendering of transparent objects
settings.use_depth_peeling = True

# Generate a torus and assign a linewidth of 1
tor = Torus(res=9).linewidth(1)

# Generate an array of random RGBA color values for each cell of the mesh
rgba = np.random.rand(tor.ncells, 4)*255
tor.cellcolors = rgba

# Print information about the cell arrays of the mesh and their shape
printc(
    'Mesh cell arrays:', tor.celldata.keys(),
    'shape:', tor.celldata['CellsRGBA'].shape,
)

# Display the mesh with the assigned colors and the docstring
show(tor, __doc__).close()



"""Find cells within specified bounds in x, y and/or z."""
from vedo import *

# Load a mesh of a shark and normalize it
mesh = Mesh(dataurl+'shark.ply').normalize()

# Set the color of the mesh and the line width to 1
mesh.color('aqua').linewidth(1)

# Define the lower and upper bounds for the z-axis
z1, z2 = -1.5, -1.0

# Find the cell IDs of cells within the z-axis bounds
ids = mesh.find_cells_in(zbounds=(z1,z2))

# Print the cell IDs in green to the console
printc('IDs of cells within bounds:\n', ids, c='g')

# Create two Plane objects at the specified z-positions
p1 = Plane(normal=(0,0,1), s=[2,2]).z(z1)
p2 = p1.clone().z(z2)

# Show the mesh, the two planes, the docstring
show(mesh, p1, p2, __doc__, axes=9).close()



"""Mouse hind limb growth from day 10 9h to day 15 21h"""
from vedo import settings, dataurl, load
from vedo import Text2D, Plotter, Picture, Axes, Line


def sliderfunc(widget, event):
    i = int(widget.value)
    days = int((i * 2 + 249) / 24)
    widget.title = f"{days}d {i*2+249-days*24}h"
    # remove the old and add the new shape
    # (no need to render as the slider makes a call to rendering)
    plt.pop().add(objs[i])


objs = load(dataurl + "timecourse1d.npy")  # load a list of shapes

settings.default_font = "Glasgo"

plt = Plotter(bg="blackboard")
plt += Text2D(__doc__, pos="top-center", s=1.2, c="w")
plt += Picture(dataurl + "images/limbs_tc.jpg").scale(0.0154).y(10)
plt += Line([(0, 8), (0, 10), (28.6, 10), (4.5, 8)], c="gray")
plt += Axes(objs[-1])
plt += objs[0]
plt.add_slider(
    sliderfunc,
    0,
    len(objs) - 1,
    pos=[(0.4, 0.1), (0.9, 0.1)],
    show_value=False,
    title_size=1.5,
)
plt.show(zoom=1.2, mode="image")
plt.close()



"""Fit spheres to a region of a surface defined by
N points that are closest to a given point of the surface.
For some of these point we show the fitting sphere.
Red lines join the center of the sphere to the surface point.
Blue points are the N points used for fitting"""
from vedo import *

settings.default_font = 'Kanopus'
settings.use_depth_peeling = True

plt = Plotter()

# load mesh and increase by a lot subdivide(2) the nr of surface vertices
cow = Mesh(dataurl+"cow.vtk").alpha(0.3).subdivide(2)

for i, p in enumerate(cow.points()):
    if i % 1000:
        continue  # skip most points
    pts = cow.closest_point(p, n=16)   # find the n-closest points to p
    sph = fit_sphere(pts).alpha(0.05)  # find the fitting sphere
    if sph is None:
        continue  # may fail if all points sit on a plane
    plt += sph
    plt += Points(pts)
    plt += Line(sph.center, p, lw=2)

plt += [cow, __doc__]
plt.show(viewup="z", axes=1).close()



"""Use a variant of the Moving Least Squares (MLS)
algorithm to project a cloud of points to become a smooth surface.
In the second window we show the error estimated for
each point in color scale (left) or in size scale (right)."""
from vedo import *
printc(__doc__, invert=1)

plt1 = Plotter(N=3, axes=1)

mesh = Mesh(dataurl+"bunny.obj").normalize().subdivide()

pts = mesh.points()
pts += np.random.randn(len(pts), 3)/20  # add noise, will not mess up the original points


#################################### smooth cloud with MLS
# build the mesh points
s0 = Points(pts, r=3).color("blue")
plt1.at(0).show(s0, "original point cloud + noise")

# project s1 points into a smooth surface of points
# The parameter f controls the size of the local regression.
mls1 = s0.clone().smooth_mls_2d(f=0.5)
plt1.at(1).show(mls1, "MLS first pass, f=0.5")

# mls1 is an Assembly so unpack it to get the first object it contains
mls2 = mls1.clone().smooth_mls_2d(radius=0.1)
plt1.at(2).show(mls2, "MLS second pass, radius=0.1")


#################################### draw errors
plt2 = Plotter(pos=(300, 400), N=2, axes=1)

variances = mls2.info["variances"]
vmin, vmax = np.min(variances), np.max(variances)
print("min and max of variances:", vmin, vmax)
vcols = [color_map(v, "jet", vmin, vmax) for v in variances]  # scalars->colors

sp0 = Spheres(mls2.points(), c=vcols, r=0.02) # error as color
sp1 = Spheres(mls2.points(), c="red", r=variances/4) # error as point size

mesh.color("k").alpha(0.05).wireframe()

plt2.at(0).show(sp0, "Use color to represent variance")
plt2.at(1).show(sp1, "point size to represent variance", zoom=1.3, interactive=True)
plt2.close()
plt1.close()



"""Add a secondary y-axis for units conversion"""
from vedo import np, settings, dataurl, Mesh, show
from vedo.pyplot import plot, Figure

settings.annotated_cube_texts = ['front','back','left','right','top','bttm']

x0, x1 = [0.3, 2.0]
x = np.linspace(x0, x1, num=50)

# The main plot
fig1 = plot(
    x,
    1000*np.cos(x+1),
    xlim=[x0, x1],
    ylim=[-1000, 250],
    aspect=16/9,
    padding=0,                    # do not mess up with margins
    title="Wing pull vs position",
    xtitle="Distance from airplane axis [m]",
    ytitle="N [Kg*m/s^2 ]",
    axes=dict(
        xygrid_transparent=False,
        xygrid_color='k7',
        xyalpha=1,
        xyplane_color='w',
        yhighlight_zero=True,
    ),
)

# This empty Figure just creates a new y-axis in red
fig2 = Figure(
    fig1.xlim,                    # same as fig1
    fig1.ylim * 7.236,            # units conversion factor
    aspect=fig1.aspect,           # same as fig1
    padding=fig1.padding,         # same as fig1
    xtitle='',                    # don't draw the x-axis!
    ytitle='Poundal [lb*ft/s^2 ]',
    axes=dict(                    # extra options for y-axis
        number_of_divisions=10,
        yshift_along_x=1,         # shift 100% to the right
        ylabel_offset=-1,
        ylabel_justify="center-left",
        ytitle_position=0.5,
        ytitle_justify="top-center",
        axes_linewidth=3,
        c='red3',
    ),
)

fig1.rotate_x(90).rotate_z(90).shift(-0.5, 0, 1)
fig2.rotate_x(90).rotate_z(90).shift(-0.5, 0, 1)

msh = Mesh(dataurl+"cessna.vtk")

cam = dict(  # press C to get these values
    pos=(3.899, -0.4781, 1.157),
    focal_point=(-0.1324, 0.9041, 0.3530),
    viewup=(-0.1725, 0.06857, 0.9826),
)
show(msh, fig1, fig2, __doc__, axes=5, camera=cam, bg2='lb').close()





"""pymeshlab interoperability example:
Surface reconstruction by ball pivoting"""
import vedo
import pymeshlab  # tested on pymeshlab-2022.2.post2

pts = vedo.Mesh(vedo.dataurl+'cow.vtk').points() # numpy array of vertices

m = pymeshlab.Mesh(vertex_matrix=pts)

ms = pymeshlab.MeshSet()
ms.add_mesh(m)

p = pymeshlab.Percentage(2)
ms.generate_surface_reconstruction_ball_pivoting(ballradius=p)

mlab_mesh = ms.current_mesh()
reco_mesh = vedo.Mesh(mlab_mesh).compute_normals().flat().backcolor('t')

vedo.show(
    __doc__, vedo.Points(pts), reco_mesh,
    axes=True, bg2='blue9', title="vedo + pymeshlab",
)
